"""
Hierarchical resampling module for Python
=========================================

hierarch is a Python module implementing fast randomization algorithms using
numpy, numba, sympy, and scipy. 

It aims to make design-based randomization tests as easy as possible for experimental
researchers.
"""
import hierarch.stats
import hierarch.power
import hierarch.resampling
